use std::time::Instant;
use log::info;
use noise::{Fbm, MultiFractal, NoiseFn, Perlin};
use rand::prelude::*;
use voronoice::{BoundingBox, Point, Voronoi, VoronoiBuilder};
use winit::{event::{DeviceEvent, Event, KeyEvent, MouseButton, MouseScrollDelta, WindowEvent}, keyboard::{KeyCode, PhysicalKey}, window::Window};

use crate::graphics::{Uniforms, Vertex};

const FRAMERATE: usize = 1; // Update per second

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum CellKind {
    Void,
    Sea,
    Beach,
    Forest,
    Dirt,
    Stone,
    Grass
}

#[derive(Debug)]
struct CellData {
    kind: CellKind,
    cell: usize,
    z: f32,
    area: f32,
    moisture: f32, 
    ressource: (f32, usize) // How much ressource there is (between 0 and 1) and the last time it was updated (in frames)
}
impl CellData {
    fn new(kind: CellKind, cell: usize, z: f32, moisture: f32, ressource: f32, t: usize, voronoi: &Voronoi) -> Self {
        let mut area = 0.;
        let c = voronoi.cell(cell);
        let vs = c.iter_vertices().collect::<Vec<_>>();
        let a = vs[0];
        for i in 1..(vs.len()-1) {
            let b = vs[i];
            let c = vs[i+1];
            area += 0.5 * ((a.x*(b.y-c.y))+(b.x*(c.y-a.y))+(c.x*(a.y-b.y))).abs();
        }
        Self {
            kind,
            cell,
            z,
            area: area as f32,
            moisture,
            ressource: (ressource, t)
        }
    }
    fn color(&self) -> [f32; 4] {
        // let mut rng = thread_rng();
        // [rng.gen(), rng.gen(), rng.gen(), 1.]
        match self.kind {
            CellKind::Void => [0.; 4],
            CellKind::Sea => [0., 0., 1., 1.],
            CellKind::Beach => [0.82, 0.84, 0.51, 1.],
            CellKind::Forest => [0., 0.5 - (self.moisture*0.4), 0., 1.],
            CellKind::Dirt => [0.53 - (self.moisture*0.4), 0.38-(self.moisture*0.4), 0.29-(self.moisture*0.4), 1.],
            CellKind::Stone => [0.5, 0.5, 0.5, 1.],
            CellKind::Grass => [(136./255.) - (self.moisture*0.4), (204./255.) - (self.moisture*0.4), (59./255.) - (self.moisture*0.4), 1.]
        }
    }
    fn ressource(&self) -> f32 {
        // How much it get by day
        let recuperation_rate = match self.kind {
            CellKind::Void | CellKind::Sea | CellKind::Beach | CellKind::Dirt | CellKind::Stone => 0.,
            CellKind::Forest => 1. / (100. * 365.25), // Let's say that a forest takes 100 years to mature
            CellKind::Grass => 1. / (7. * 7.) // Let's say that grass takes 7 weaks to reach its max
        };
        self.ressource.0 + (self.ressource.1 as f32 * recuperation_rate)
    }
    fn set_ressource(&mut self, val: f32, t: usize) {
        assert!(val >= 0. && val <= 1.);
        self.ressource = (val, t);
    }
    fn refresh_ressource(&mut self, t: usize) {
        self.ressource = (self.ressource(), t);
    }
}

struct Map {
    voronoi: Voronoi,
    cells_data: Vec<CellData>,
    seed: u32
}
impl Map {
    const HEIGHT: f32 = 2.;
    const WIDTH: f32 = 2.;
    const REAL_HEIGHT: f32 = 500.;
    const REAL_WIDTH: f32 = 500.;
    const SIZE: usize = 10_000;
    fn new(seed: u32, t: usize) -> Self {
        let mut rng = rand::rngs::SmallRng::seed_from_u64(seed as u64);
        let mut sites = Vec::with_capacity(Self::SIZE);
        for _ in 0..Self::SIZE {
            sites.push(Point { x:rng.gen_range(-Self::WIDTH/2.0..Self::WIDTH/2.0) as f64, y:rng.gen_range(-Self::HEIGHT/2.0..Self::HEIGHT/2.0) as f64 })
        }
        let voronoi = VoronoiBuilder::default()
            .set_sites(sites)
            .set_bounding_box(BoundingBox::new_centered(Self::WIDTH as f64, Self::HEIGHT as f64))
            .set_lloyd_relaxation_iterations(3)
            .build()
            .unwrap();
        let mut cells_data = Vec::with_capacity(Self::SIZE);
        let z_noise = Fbm::<Perlin>::new(seed);
        let moisture_noise = Fbm::<Perlin>::new(seed+1)
            .set_frequency(2.);
        for i in 0..Self::SIZE {
            let c = voronoi.cell(i);
            let site = c.site_position();
            let z = (
                0.3 // Arbitrary value
                + ((z_noise.get([site.x, site.y])+1.)/2.) // Noise + [0; 1]
                - ((site.x.powi(2)+site.y.powi(2)).sqrt()*0.5) // Distance - [0; sqrt(2)] * 0.5
            ).clamp(0., 1.);
            let m = (
                (moisture_noise.get([site.x, site.y])+1.)/2. // Noise + [0; 1]
            ).clamp(0., 1.) as f32;
            let k = if z <= 0.5 {
                CellKind::Sea
            } else if z <= 0.52 {
                CellKind::Beach
            } else if z < 0.8 {
                CellKind::Dirt
            } else {
                CellKind::Stone
            };
            cells_data.push(CellData::new(k, i, z as f32, m, 1., t, &voronoi));
        }
        Self {
            voronoi,
            cells_data,
            seed
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum EntityKind {
    Horse
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum EntityState {
    Walking(Instant), // Start of walk
    Resting
}

#[derive(Debug)]
struct Entity {
    cell: usize,
    kind: EntityKind,
    start: Instant,
    state: EntityState,
    health: f32 // between 0 and 1
}
impl Entity {
    fn new(cell: usize, kind: EntityKind) -> Self {
        Self {
            cell,
            kind,
            start: Instant::now(),
            state: EntityState::Resting,
            health: 1.
        }
    }
    fn pos<'a>(&self, map: &'a Map) -> &'a Point {
        &map.voronoi.sites()[self.cell]
    }
    fn render(&self, vertices: &mut Vec<Vertex>, indices: &mut Vec<u32>, map: &Map) {
        match self.kind {
            EntityKind::Horse => {
                let color = [171./255., 122./255., 50./255., 1.];
                let dark = [color[0]-0.3, color[1]-0.3, color[2]-0.3, 1.];
                let (vs, is) = match self.state {
                    EntityState::Walking(now) => {
                        let now = now.elapsed().as_secs_f32()*5.;
                        (
                            [
                                // back left leg
                                Vertex { pos: [-0.5, 0.3], color: dark },
                                Vertex { pos: [-0.4 + (now.sin()*0.1), -0.7 + (now.cos().max(-0.5)*0.1)], color: dark },
                                Vertex { pos: [-0.25, 0.1], color: dark },

                                // back right leg
                                Vertex { pos: [-0.5, 0.3], color },
                                Vertex { pos: [-0.4 + ((now + 1.).sin()*0.1), -0.7 + ((now + 1.).cos().max(-0.5)*0.1)], color },
                                Vertex { pos: [-0.25, 0.1], color },

                                // front left leg
                                Vertex { pos: [0.3, 0.2], color: dark },
                                Vertex { pos: [0.4 + ((now-1.).sin()*0.1), -0.7 + ((now-1.).cos().max(-0.5)*0.1)], color: dark },
                                Vertex { pos: [0.5, 0.3], color: dark },

                                // front right leg
                                Vertex { pos: [0.3, 0.2], color },
                                Vertex { pos: [0.4 + ((now-2.).sin()*0.1), -0.7 + ((now-2.).cos().max(-0.5)*0.1)], color },
                                Vertex { pos: [0.5, 0.3], color },

                                // body
                                // 3
                                Vertex { pos: [-0.3, 0.], color },
                                Vertex { pos: [0.4, -0.1], color },
                                // 11
                                Vertex { pos: [0.3, 0.4], color },
                            ],
                            [
                                0,1,2,
                                3,4,5,
                                6,7,8,
                                9,10,11,
                                3,12,13,
                                3,13,11,
                                3,11,14
                            ]
                        )
                    },
                    EntityState::Resting => {
                        (
                            [
                                // back left leg
                                Vertex { pos: [-0.5, 0.3], color: dark },
                                Vertex { pos: [-0.4, -0.75], color: dark },
                                Vertex { pos: [-0.25, 0.1], color: dark },

                                // back right leg
                                Vertex { pos: [-0.5, 0.3], color },
                                Vertex { pos: [-0.4, -0.75], color },
                                Vertex { pos: [-0.25, 0.1], color },

                                // front left leg
                                Vertex { pos: [0.3, 0.2], color: dark },
                                Vertex { pos: [0.4, -0.75], color: dark },
                                Vertex { pos: [0.5, 0.3], color: dark },

                                // front right leg
                                Vertex { pos: [0.3, 0.2], color },
                                Vertex { pos: [0.4, -0.75], color },
                                Vertex { pos: [0.5, 0.3], color },

                                // body
                                // 3
                                Vertex { pos: [-0.3, 0.], color },
                                Vertex { pos: [0.4, -0.1], color },
                                // 11
                                Vertex { pos: [0.3, 0.4], color },
                            ],
                            [
                                0,1,2,
                                3,4,5,
                                6,7,8,
                                9,10,11,
                                3,12,13,
                                3,13,11,
                                3,11,14
                            ]
                        )
                    }
                };

                let pos = self.pos(map);

                vertices.reserve(vs.len());
                let base = vertices.len() as u32;
                for mut v in vs {
                    v.pos[0] = v.pos[0]/50. + pos.x as f32;
                    v.pos[1] = (v.pos[1] + 0.75)/50. + pos.y as f32;
                    vertices.push(v)
                }

                indices.reserve(is.len());
                for i in is {
                    indices.push(base + i);
                }
            }
        }
    }
    fn update(&mut self, map: &mut Map) {
        // Let’s take 0.57 kg of grass / m2
        // Let’s take 7.5 kg of food / day for a horse
        let r = dbg!(&map.cells_data[self.cell]).ressource();
        
    }
}

pub struct State {
    pub vertices: Vec<Vertex>,
    pub indices: Vec<u32>,
    pub uniforms: Uniforms,
    map: Map,
    entities: Vec<Entity>,
    start: Instant,
    t: usize, // Time in frames
    selected_tile: usize,
    mouse_pressed: bool
}
impl State {
    pub fn new() -> Self {
        let mut s = Self {
            vertices: vec![],
            indices: vec![],
            uniforms: Uniforms::default(),
            start: Instant::now(),
            t: 0,
            map: Map::new(0, 0),
            entities: Vec::new(),
            selected_tile: 0,
            mouse_pressed: false
        };
        s.render();
        s
    }
    pub fn event(&mut self, event: &Event<()>, window: &Window) {
        match event {
            Event::WindowEvent { event: WindowEvent::MouseInput { state, button, ..}, ..} => {
                if state.is_pressed() {
                    match button {
                        MouseButton::Left => {
                            self.entities.push(Entity::new(self.selected_tile, EntityKind::Horse));
                            self.mouse_pressed = true;
                        },
                        MouseButton::Right => self.map = Map::new(self.map.seed + 1, self.t),
                        _ => {}
                    };
                } else {
                    match button {
                        MouseButton::Left => {
                            self.mouse_pressed = false;
                        },
                        _ => {}
                    }
                }
            },
            Event::WindowEvent { event: WindowEvent::CursorMoved { position, .. }, ..} => {
                let w_size = window.inner_size();
                let pos = Point {
                    x: ((2.*position.x/(w_size.width as f64))-1.)*(self.uniforms.camera[2] as f64) + self.uniforms.camera[0] as f64,
                    y: (1.-(2.*position.y/(w_size.height as f64)))*(self.uniforms.camera[2] as f64) + self.uniforms.camera[1] as f64
                };
                let c = self.map.voronoi.cell(self.selected_tile);
                if self.mouse_pressed {
                    for i in c.iter_path(pos.clone()) {
                        self.selected_tile = i;
                        let cd = &mut self.map.cells_data[self.selected_tile];
                        if let CellKind::Dirt = cd.kind {
                            cd.kind = CellKind::Forest;
                        }
                    }
                } else {
                    self.selected_tile = c.iter_path(pos).last().unwrap();
                }
            },
            Event::DeviceEvent { event: DeviceEvent::MouseWheel { delta }, ..} => {
                self.uniforms.camera[2] -= match delta {
                    MouseScrollDelta::PixelDelta(pos) => pos.y as f32,
                    MouseScrollDelta::LineDelta(_, y) => *y
                };
                self.uniforms.camera[2] = self.uniforms.camera[2].clamp(0.1, 1.);
            },
            Event::WindowEvent { event: WindowEvent::KeyboardInput { event: KeyEvent { physical_key: PhysicalKey::Code(kc), state, .. }, .. }, .. } => {
                if state.is_pressed() {
                    match kc {
                        KeyCode::KeyW => {
                            self.uniforms.camera[1] += 0.1;
                        },
                        KeyCode::KeyS => {
                            self.uniforms.camera[1] -= 0.1;
                        },
                        KeyCode::KeyA => {
                            self.uniforms.camera[0] -= 0.1;
                        },
                        KeyCode::KeyD => {
                            self.uniforms.camera[0] += 0.1;
                        },
                        _ => {}
                    }
                }
            },
            _ => {}
        }
    }
    pub fn render(&mut self) {
        info!("render");
        self.vertices = Vec::new();
        self.indices = Vec::new();

        for (c, cd) in self.map.voronoi.iter_cells().zip(self.map.cells_data.iter()).filter(|(_,cd)| cd.kind != CellKind::Forest) {
            let mut color = cd.color();
            if c.site() == self.selected_tile {
                color[0] = (color[0]+0.4).clamp(0., 1.);
                color[1] = (color[1]+0.4).clamp(0., 1.);
                color[2] = (color[2]+0.4).clamp(0., 1.);
            }
            let vs = c.iter_vertices().collect::<Vec<_>>();
            let i = self.vertices.len() as u32;
            for v in vs.iter() {
                self.vertices.push(Vertex { pos: [v.x as f32, v.y as f32], color });
            }
            for v in 1..(vs.len()-1) as u32 {
                self.indices.push(i);
                self.indices.push(i+v);
                self.indices.push(i+v+1);
            }
        }

        for e in self.entities.iter() {
            e.render(&mut self.vertices, &mut self.indices, &self.map);
        }

        for (c, cd) in self.map.voronoi.iter_cells().zip(self.map.cells_data.iter()).filter(|(_,cd)| cd.kind == CellKind::Forest) {
            let mut color = cd.color();
            if c.site() == self.selected_tile {
                color[0] = (color[0]+0.4).clamp(0., 1.);
                color[1] = (color[1]+0.4).clamp(0., 1.);
                color[2] = (color[2]+0.4).clamp(0., 1.);
            }
            let vs = c.iter_vertices().collect::<Vec<_>>();
            let i = self.vertices.len() as u32;
            for v in vs.iter() {
                self.vertices.push(Vertex { pos: [v.x as f32, v.y as f32], color });
            }
            for v in 1..(vs.len()-1) as u32 {
                self.indices.push(i);
                self.indices.push(i+v);
                self.indices.push(i+v+1);
            }
        }
    }
    pub fn update_if_needed(&mut self) {
        while self.start.elapsed().as_secs_f32() > (self.t as f32) / (FRAMERATE as f32) {
            self.update();
        }
    }
    pub fn update(&mut self) {
        info!("update");
        let mut rng = thread_rng();
        let mut new_kind = Vec::new();
        for cd in self.map.cells_data.iter() {
            if cd.kind == CellKind::Forest || cd.kind == CellKind::Grass {
                let r = rng.gen::<f32>();
                if r < (0.035*cd.moisture) {
                    let c = self.map.voronoi.cell(cd.cell);
                    let n = c.iter_neighbors().choose(&mut rng).unwrap();
                    let k = if r < (0.005*cd.moisture) && (self.map.cells_data[n].kind == CellKind::Dirt || self.map.cells_data[n].kind == CellKind::Grass) && cd.kind == CellKind::Forest {
                        Some(CellKind::Forest)
                    } else if self.map.cells_data[n].kind == CellKind::Dirt {
                        Some(CellKind::Grass)
                    } else {
                        None
                    };
                    if let Some(k) = k {
                        new_kind.push((n, k));
                    }
                }
            }
        }
        for (n, k) in new_kind {
            self.map.cells_data[n].kind = k;
        }

        for e in self.entities.iter_mut() {
            e.update(&mut self.map);
        }

        self.t += 1;
    }
}